'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.stringify = exports.createAssignToObject = exports.getStandardValue = exports.getRelativeValue = exports.getKeysMetadata = exports.getCrioValue = exports.getCorrectConstructor = exports.isReactElement = exports.isEqual = exports.isCrioArray = exports.isCrio = exports.isComplexObject = exports.freeze = undefined;

var _deepFreezeStrict = require('deep-freeze-strict');

var _deepFreezeStrict2 = _interopRequireDefault(_deepFreezeStrict);

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _isPlainObject = require('lodash/isPlainObject');

var _isPlainObject2 = _interopRequireDefault(_isPlainObject);

var _stringifier = require('stringifier');

var _stringifier2 = _interopRequireDefault(_stringifier);

var _constants = require('./constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var STRINGIFIER_OPTIONS = {
  maxDepth: 10
};

/**
 * @private
 *
 * @function freeze
 *
 * @description
 * freeze the object if it is production
 *
 * @param {CrioArray|CrioObject} crio object to freeze
 *
 * @returns {T} frozen object
 */


// constants
// external dependencies
var freeze = exports.freeze = function (isProduction) {
  if (isProduction) {
    return function (object) {
      return object;
    };
  }

  return function (object) {
    return (0, _deepFreezeStrict2.default)(object);
  };
}(_constants.IS_PRODUCTION);

/**
 * @private
 *
 * @function isComplexObject
 *
 * @description
 * is the object an array or plain object
 *
 * @param  {*} object object to test
 *
 * @returns {boolean} is the object a complex object
 */
var isComplexObject = exports.isComplexObject = function isComplexObject(object) {
  return (0, _isArray2.default)(object) || (0, _isPlainObject2.default)(object);
};

/**
 * @private
 *
 * @function isCrio
 *
 * @description
 * is the object a crio object or not
 *
 * @param {*} object object to test
 *
 * @returns {boolean} is the object a crio
 */
var isCrio = exports.isCrio = function isCrio(object) {
  return !!(object && object[_constants.CRIO_SYMBOL]);
};

/**
 * @private
 *
 * @function isCrioArray
 *
 * @description
 * is the object a crio array
 *
 * @param {*} object object to test
 * @returns {boolean} is the object a crio array
 */
var isCrioArray = exports.isCrioArray = function isCrioArray(object) {
  return isCrio && object.isArray();
};

/**
 * @private
 *
 * @function isEqual
 *
 * @description
 * are the crio objects equal
 *
 * @param {CrioArray|CrioObject} crio crio object to test against
 * @param {*} object object to test equality with crio object for
 *
 * @returns {boolean} are the objects equal
 */
var isEqual = exports.isEqual = function isEqual(crio, object) {
  return isCrio(object) && crio.hashCode === object.hashCode;
};

/**
 * @private
 *
 * @function isReactElement
 *
 * @description
 * is the object passed a react element
 *
 * @param {*} object object to test
 * @returns {boolean} is object a react element
 */
var isReactElement = exports.isReactElement = function isReactElement(object) {
  return !!object && object.$$typeof === _constants.REACT_ELEMENT_TYPE;
};

/**
 * @private
 *
 * @function getCorrectConstructor
 *
 * @description
 * get the constructor that is valid for the object type passed
 *
 * @param {*} object object to test
 * @param {CrioArray} CrioArray constructor for CrioArray class
 * @param {CrioObject} CrioObject constructor for CrioObject class
 * @returns {CrioArray|CrioObject} constructor correct for object
 */
var getCorrectConstructor = exports.getCorrectConstructor = function getCorrectConstructor(object, CrioArray, CrioObject) {
  return (0, _isArray2.default)(object) ? CrioArray : CrioObject;
};

/**
 * @private
 *
 * @function getCrioValue
 *
 * @description
 * get the value based on its type
 *
 * @param {*} object object to get value of
 * @param {Function} Constructor function to call new of if object is complex
 *
 * @returns {*} object with clean value
 */
var getCrioValue = exports.getCrioValue = function getCrioValue(object, Constructor) {
  if (isCrio(object) || isReactElement(object)) {
    return object;
  }

  if (isComplexObject(object)) {
    return new Constructor(object);
  }

  return object;
};

/**
 * @private
 *
 * @function getKeysMetadata
 *
 * @description
 * get the value at the parent key location
 *
 * @param {Array<number|string>} keys keys to get value of
 * @param {CrioArray|CrioObject} instance crio instance
 *
 * @returns {{currentValue: *, lastIndex: number}} parent key metadata
 */
var getKeysMetadata = exports.getKeysMetadata = function getKeysMetadata(keys, instance) {
  var lastIndex = keys.length - 1;
  var parentKeys = keys.slice(0, lastIndex);
  var currentValue = instance.getIn(parentKeys);

  return {
    currentValue: currentValue,
    lastIndex: lastIndex,
    parentKeys: parentKeys
  };
};

/**
 * @private
 *
 * @function getRelativeValue
 *
 * @description
 * get the relative value used in copyWithin
 *
 * @param {number} value value used as baseline
 * @param {number} length the length of the crio
 * @returns {number} the relative number value
 */
var getRelativeValue = exports.getRelativeValue = function getRelativeValue(value, length) {
  return value < 0 ? Math.max(length + value, 0) : Math.min(value, length);
};

/**
 * @private
 *
 * @function getStandardValue
 *
 * @description
 * get the standard value (thawed if crio)
 *
 * @param {*} object object to get standard version of
 *
 * @returns {*} standard version of object
 */
var getStandardValue = exports.getStandardValue = function getStandardValue(object) {
  return isCrio(object) ? object.thaw() : object;
};

/**
 * @private
 *
 * @function createAssignToObject
 *
 * @description
 * create a function that will assign a value to an object
 *
 * @param {Function} CrioArray constructor for crio arrays
 * @param {Function} CrioObject constructor for crio objects
 *
 * @returns {function((Array<*>|Object), function): function(*, string): void} assignment function
 */
var createAssignToObject = exports.createAssignToObject = function createAssignToObject(CrioArray, CrioObject) {
  return function (object, getValue) {
    return function (value, key) {
      object[key] = getValue(value, getCorrectConstructor(value, CrioArray, CrioObject));
    };
  };
};

/**
 * @private
 *
 * @function stringify
 *
 * @description
 * convert object to string
 *
 * @param {*} object object to stringify
 *
 * @returns {string} stringified version of object
 */
var stringify = exports.stringify = (0, _stringifier2.default)(STRINGIFIER_OPTIONS);